# -*- coding: utf-8 -*-

from ..species import species_dict, _nth_digit, _normalize_species


def test_species():
    num = 9876543210

    assert [_nth_digit(num, n) for n in range(10)] == list(range(10)), \
        'Incorrect digits extracted from {}'.format(num)

    num = 7492

    assert [_nth_digit(num, n) for n in range(5, -1, -1)] == \
        [0, 0, 7, 4, 9, 2], \
        'Incorrect digits extracted from {}'.format(num)

    assert all(i > 100 for i in species_dict), \
        'Elementary particles in species data.'

    assert len(set(species_dict)) == len(species_dict), \
        'Duplicates in species data.'

    pion_data = species_dict[211]

    assert pion_data['name'] == 'pi', \
        'Incorrect pion name.'

    assert abs(pion_data['mass'] - .13957061) < 1e-15, \
        'Incorrect pion mass.'

    assert pion_data['has_anti'], \
        'The pi+ has an antiparticle.'

    assert pion_data['charge'] == 1, \
        'The pi+ has +1 charge.'

    assert pion_data['boson'], \
        'The pion is a boson.'

    assert not species_dict[111]['has_anti'], \
        'The pi0 does not have an antiparticle.'

    assert species_dict[311]['has_anti'], \
        'The K0 has an antiparticle.'

    assert not species_dict[2212]['boson'], \
        'The proton is a fermion.'

    assert 'mass_range' not in species_dict[211], \
        'The pion is stable.'

    assert 'mass_range' in species_dict[213], \
        'The rho is unstable.'

    assert abs(species_dict[213]['mass_range'][0] - .28) < 1e-12, \
        'The rho mass threshold is two pions.'

    assert all(
        i['mass_range'][0] >= .28
        for i in species_dict.values() if 'mass_range' in i
    ), 'The lightest decay product is two pions.'

    normalized = _normalize_species()

    assert len(normalized) == sum(
        2 if info['has_anti'] else 1 for info in species_dict.values()
    ), 'Incorrect number of normalized species.'

    assert normalized[0][0] == 111, \
        'First normalized species should be the pi0.'

    assert normalized[1][0] == -normalized[2][0] == 211, \
        'Second and third normalized species should be the pi+/-.'

    assert normalized[1][1] is normalized[2][1], \
        'pi+/- should share the same info dict.'

    normalized_id = _normalize_species('id')

    assert [i[0] for i in normalized_id] == [
        211, -211, 321, -321, 2212, -2212
    ], 'Incorrect normalized species IDs.'

    test_ID = 2112
    test_info = species_dict[test_ID]
    assert _normalize_species([test_ID]) == [
        (test_ID, test_info),
        (-test_ID, test_info)
    ], 'Incorrect custom normalized species.'



# -*- coding: utf-8 -*-

import numpy as np

from .. import Surface, HRG, sample


def test_dtype():
    surface = Surface([1, 0, 0], [1, 0, 0], [0, 0])
    hrg = HRG(.15)

    parts = sample(surface, hrg)

    dt = parts.dtype

    assert dt.names == ('ID', 'x', 'p'), 'Incorrect dtype fields.'

    assert dt['ID'].shape == (), 'Incorrect ID field shape.'
    assert dt['x'].shape == (4,), 'Incorrect x field shape.'
    assert dt['p'].shape == (4,), 'Incorrect x field shape.'


def test_ymax():
    hrg = HRG(.15)

    ymax = np.random.uniform(.1, 1.)

    surface = Surface(
        np.array([[1., 0., 0.]]),
        np.array([[1e3/hrg.density(), 0., 0.]]),
        np.array([[0., 0.]]),
        ymax=ymax
    )

    parts = sample(surface, hrg)
    p = parts['p']

    E, px, py, pz = p.T
    y = .5*np.log((E + pz)/(E - pz))

    assert np.all(np.fabs(y) < ymax), 'Rapidity outside ymax.'


def test_decay_f500():
    hrg = HRG(.15, species=[9000221], decay_f500=True)

    surface = Surface(
        np.array([[1., 0., 0.]]),
        np.array([[1e2/hrg.density(), 0., 0.]]),
        np.array([[0., 0.]]),
    )

    parts = sample(surface, hrg)
    i = parts['ID']

    assert np.all((i == 111) | (i == 211) | (i == -211)), \
        'f500 not decayed to pions.'



# -*- coding: utf-8 -*-

import pkgutil

__all__ = ['species_dict']


def _nth_digit(i, n):
    """
    Determine the nth digit of an integer, starting at the right and counting
    from zero.

    >>> [_nth_digit(123, n) for n in range(5)]
    [3, 2, 1, 0, 0]

    """
    return (i // 10**n) % 10


def _mass_range(name, mass, width, degen, boson):
    """
    Determine the mass thresholds of a resonance for truncating its
    Breit-Wigner distribution.

    """
    # Masses of decay products, rounded up so that the minimum threshold is
    # slightly larger than the actual total mass of the decay products.
    PION = .14
    KAON = .50
    ETA = .55
    NUCLEON = .94
    LAMBDA = 1.12
    SIGMA = 1.20
    XI = 1.33

    if boson:  # meson
        m_min = {
            'f(0)(500)': 2*PION,
            'rho(770)': 2*PION,
            'omega(782)': 2*PION,
            "eta'(958)": 2*PION,
            'f(0)(980)': 2*PION,
            'a(0)(980)': ETA + PION,
            'phi(1020)': 3*PION,
            'h(1)(1170)': 3*PION,
            'b(1)(1235)': 4*PION,
            'a(1)(1260)': 3*PION,
            'f(2)(1270)': 2*PION,
            'f(1)(1285)': 2*PION,
            'eta(1295)': ETA + 2*PION,
            'pi(1300)': 3*PION,
            'a(2)(1320)': 3*PION,
            'f(0)(1370)': 2*PION,
            'pi(1)(1400)': ETA + PION,
            'eta(1405)': ETA + 2*PION,
            'f(1)(1420)': 2*KAON + PION,
            'omega(1420)': 3*PION,
            'a(0)(1450)': ETA + PION,
            'rho(1450)': 2*PION,
            'eta(1475)': 2*KAON + PION,
            'f(0)(1500)': 2*PION,
            "f(2)'(1525)": 2*PION,
            'pi(1)(1600)': 3*PION,
            'eta(2)(1645)': ETA + 2*PION,
            'omega(1650)': 3*PION,
            'omega(3)(1670)': 3*PION,
            'pi(2)(1670)': 3*PION,
            'phi(1680)': 2*KAON,
            'rho(3)(1690)': 2*PION,
            'rho(1700)': 2*PION,
            'f(0)(1710)': 2*PION,
            'pi(1800)': 3*PION,
            'phi(3)(1850)': 2*KAON,
            'f(2)(1950)': 2*PION,
            'f(2)(2010)': 2*KAON,
            'a(4)(2040)': 3*PION,
            'f(4)(2050)': 2*PION,
            'f(2)(2300)': 2*KAON,
            'f(2)(2340)': 2*ETA,
            'K*(892)': KAON + PION,
            'K(1)(1270)': KAON + 2*PION,
            'K(1)(1400)': KAON + 2*PION,
            'K*(1410)': KAON + PION,
            'K(0)*(1430)': KAON + PION,
            'K(2)*(1430)': KAON + PION,
            'K*(1680)': KAON + PION,
            'K(2)(1770)': KAON + 2*PION,
            'K(3)*(1780)': KAON + PION,
            'K(2)(1820)': KAON + 2*PION,
            'K(4)*(2045)': KAON + PION,
        }.get(name)

    else:  # baryon
        m_min = {
            'N': NUCLEON + PION,
            'Delta': NUCLEON + PION,
            'Lambda': SIGMA + PION,
            'Sigma': LAMBDA + PION,
            'Xi': XI + PION,
            'Omega': XI + KAON + PION,
        }.get(name.split('(')[0])

    assert m_min is not None, \
        'unknown mass threshold for {}'.format(name)

    assert mass > m_min, \
        'minimum mass larger than pole mass for {}'.format(name)

    # Truncate the distribution at several widths.  The Breit-Wigner shape has
    # a long, slowly-decaying tail so we must cut it somewhere.
    m_min = max(m_min, mass - 4*width)
    m_max = mass + 4*width

    return m_min, m_max


def _read_particle_data():
    """
    Parse particle data from the PDG table.

    Yields pairs (ID, data) where ID is the Monte Carlo ID number and data is a
    dict of properties.

    """
    charge_codes = {
        b'-' : -1,
        b'0' :  0,
        b'+' :  1,
        b'++':  2,
    }

    for l in pkgutil.get_data('frzout', 'mass_width_2017.mcd').splitlines():
        # skip comments
        if l.startswith(b'*'):
            continue

        # extract particle IDs (possibly multiple on a line)
        IDs = [int(i) for i in l[:32].split()]

        # skip elementary particles
        if IDs[0] < 100:
            continue

        # digits 1-3 of the ID denote the quark content
        q1, q2, q3 = quarks = [_nth_digit(IDs[0], n) for n in [1, 2, 3]]

        # skip charm and bottom particles
        if any(q > 3 for q in quarks):
            continue

        # the last digit (ones place) of the ID is the degeneracy
        degen = _nth_digit(IDs[0], 0)

        # skip K0S (ID = 310) and K0L (130)
        if degen == 0:
            continue

        # extract rest of data
        mass = float(l[33:50])
        try:
            width = float(l[70:87])
        except ValueError:
            width = 0.
        name, charges = l[107:].split()
        name = name.strip().decode()
        charges = [charge_codes[i] for i in charges.split(b',')]

        assert len(IDs) == len(charges)

        base_data = dict(
            name=name,
            mass=mass,
            width=width,
            degen=degen,
            boson=bool(degen % 2),
        )

        # determine mass thresholds for resonances
        if width > 1e-3:
            base_data.update(mass_range=_mass_range(**base_data))

        # yield an entry for each (ID, charge) pair
        for ID, charge in zip(IDs, charges):
            data = base_data.copy()
            # The PDG does not explicitly list antiparticles.  A particle has a
            # corresponding antiparticle if any of the following are true:
            #   - it is a baryon (i.e. has three quarks)
            #   - it is charged
            #   - it is a neutral meson with two different quarks
            data.update(
                charge=charge,
                has_anti=(q3 != 0 or charge != 0 or q1 != q2),
            )
            yield ID, data


species_dict = dict(_read_particle_data())

identified = [211, 321, 2212]

urqmd = [
    2112, 2212, 12112, 12212, 1214, 2124, 22112, 22212, 32112, 32212, 2116,
    2216, 12116, 12216, 21214, 22124, 42112, 42212, 31214, 32124, 1218, 2128,
    1114, 2114, 2214, 2224, 31114, 32114, 32214, 32224, 1112, 1212, 2122, 2222,
    11114, 12114, 12214, 12224, 1116, 1216, 2126, 2226, 21112, 21212, 22122,
    22222, 21114, 22114, 22214, 22224, 11116, 11216, 12126, 12226, 1118, 2118,
    2218, 2228, 3122, 13122, 3124, 23122, 33122, 13124, 43122, 53122, 3126,
    13126, 23124, 3128, 23126, 3112, 3212, 3222, 3114, 3214, 3224, 13112,
    13212, 13222, 13114, 13214, 13224, 23112, 23212, 23222, 3116, 3216, 3226,
    13116, 13216, 13226, 23114, 23214, 23224, 3118, 3218, 3228, 3312, 3322,
    3314, 3324, 203312, 203322, 13314, 13324, 103316, 103326, 203316, 203326,
    3334, 111, 211, 221, 223, 113, 213, 9010221, 311, 321, 331, 313, 323, 333,
    10311, 10321, 9000111, 9000211, 10221, 10313, 10323, 20113, 20213, 20223,
    315, 325, 115, 215, 225, 335, 20313, 20323, 10113, 10213, 10223, 100313,
    100323, 100113, 100213, 100223, 100333, 30313, 30323, 30113, 30213, 30223,
    337, 9000221
]


def _normalize_species(species='all'):
    """
    Add antiparticles to a list of species and sort by mass.

    The single argument `species` may be:

        - 'all' (default) for all known species
        - 'id' for identified particles (pions, kaons, protons)
        - an iterable of IDs

    """
    def species_items():
        if species == 'all':
            return species_dict.items()
        elif species == 'id':
            return ((i, species_dict[i]) for i in identified)
        elif species == 'urqmd':
            return ((i, species_dict[i]) for i in urqmd)
        else:
            def items_gen():
                for ID in species:
                    info = species_dict.get(ID)
                    if info is None:
                        raise ValueError('unknown species ID: {}'.format(ID))
                    yield ID, info
            return items_gen()

    def all_species_items():
        for ID, info in species_items():
            yield ID, info
            if info['has_anti']:
                yield -ID, info

    return sorted(all_species_items(), key=lambda i: i[1]['mass'])



# -*- coding: utf-8 -*-

import numpy as np

from nose.tools import (
    assert_almost_equal,
    assert_warns_regex,
    assert_raises_regex,
)

from .. import Surface


def test_surface():
    volume = np.random.uniform(10, 100)
    tau = np.random.uniform(.5, 5.)

    x = [tau, 0, 0]
    sigma = [volume/tau, 0, 0]
    v = [0, 0]

    surf = Surface(x, sigma, v)

    assert surf.boost_invariant, 'Surface should be boost-invariant.'

    assert_almost_equal(
        surf.volume, volume, delta=1e-12,
        msg='incorrect volume'
    )

    ymax = np.random.uniform(.5, 2.)
    surf = Surface(x, sigma, v, ymax=ymax)

    assert_almost_equal(
        surf.volume, 2*ymax*volume, delta=1e-12,
        msg='incorrect volume'
    )

    x = np.random.uniform(0, 10, size=(1, 4))
    # ensure sigma is timelike (sigma^2 > 0) so that the volume is positive
    sigma = np.random.uniform([3, -1, -1, -1], [4, 1, 1, 1], size=(1, 4))
    v = np.random.uniform(-.5, .5, size=(1, 3))

    surf = Surface(x, sigma, v)

    assert not surf.boost_invariant, 'Surface should not be boost-invariant.'

    u = np.insert(v, 0, 1) / np.sqrt(1 - (v*v).sum())
    volume = np.inner(sigma, u)

    assert_almost_equal(
        surf.volume, volume, delta=1e-12,
        msg='incorrect volume'
    )

    with assert_warns_regex(Warning, 'ymax has no effect for a 3D surface'):
        Surface(x, sigma, v, ymax=np.random.rand())

    with assert_warns_regex(Warning, 'total freeze-out volume is negative'):
        Surface(x, np.concatenate([[[0]], -v], axis=1), v)

    with assert_raises_regex(
            ValueError,
            'number of spacetime dimensions of x, sigma, and/or v'
    ):
        Surface([1, 0], [1, 0], 0)

    with assert_raises_regex(
            ValueError,
            'number of spacetime dimensions of x, sigma, and/or v'
    ):
        Surface(
            np.ones((1, 4)),
            np.ones((1, 3)),
            np.ones((1, 2)),
        )

    with assert_raises_regex(
            ValueError,
            'number of elements of x, sigma, and/or v do not match'
    ):
        Surface(
            np.ones((2, 4)),
            np.ones((2, 4)),
            np.ones((3, 3)),
        )

    with assert_raises_regex(
            ValueError,
            'number of elements of pi components do not match'
    ):
        Surface(
            np.ones((3, 4)),
            np.ones((3, 4)),
            np.ones((3, 3)),
            pi=dict(
                xx=np.ones(3),
                yy=np.ones(3),
                xy=np.ones(3),
                xz=np.ones(3),
                yz=np.ones(4),
            )
        )

    with assert_raises_regex(
            ValueError,
            'number of elements of Pi do not match'
    ):
        Surface(
            np.ones((3, 4)),
            np.ones((3, 4)),
            np.ones((3, 3)),
            Pi=np.ones(4)
        )



# -*- coding: utf-8 -*-

import random

import numpy as np
from scipy import integrate
from scipy import special

from nose.tools import assert_almost_equal, assert_warns_regex

from .. import HRG, species_dict


hbarc = 0.1973269788


def test_hrg():
    ID, info = random.choice(list(species_dict.items()))
    m = info['mass']
    g = info['degen']
    sign = -1 if info['boson'] else 1

    prefactor = g / (2*np.pi**2*hbarc**3)
    if info['has_anti']:
        prefactor *= 2

    T = random.uniform(.13, .15)
    hrg = HRG(T, species=[ID], res_width=False)

    assert_almost_equal(
        hrg.T, T, delta=1e-15,
        msg='incorrect temperature'
    )

    n = np.arange(1, 50)
    density = prefactor * m*m*T * (
        (-sign)**(n-1)/n * special.kn(2, n*m/T)
    ).sum()

    assert_almost_equal(
        hrg.density(), density, delta=1e-12,
        msg='incorrect density'
    )

    def integrand(p):
        E = np.sqrt(m*m + p*p)
        return p*p * E / (np.exp(E/T) + sign)

    energy_density = prefactor * integrate.quad(integrand, 0, 10)[0]

    assert_almost_equal(
        hrg.energy_density(), energy_density, delta=1e-12,
        msg='incorrect energy density'
    )

    def integrand(p):
        E = np.sqrt(m*m + p*p)
        return p**4 / (3*E) / (np.exp(E/T) + sign)

    pressure = prefactor * integrate.quad(integrand, 0, 10)[0]

    assert_almost_equal(
        hrg.pressure(), pressure, delta=1e-12,
        msg='incorrect pressure'
    )

    with assert_warns_regex(Warning, 'high particlization temperature'):
        HRG(.193)


def test_bulk_corrections():
    hrg = HRG(.15, res_width=False)
    p0 = hrg.pressure()

    Pi_min, Pi_max = hrg.Pi_lim()

    assert Pi_min == -p0, \
        'The minimum bulk pressure should equal negative the ideal pressure.'

    assert 0 < Pi_max <= p0, \
        'The maximum bulk pressure should be <= the ideal pressure.'

    assert hrg.bulk_scale_factors(0) == (1, 1), \
        'The scale factors should equal one at zero bulk pressure.'

    nscale, pscale = hrg.bulk_scale_factors(random.uniform(.1, .9)*Pi_max)
    assert nscale < 1 and pscale > 1, \
        'Incorrect scale factors for positive bulk pressure.'

    nscale, pscale = hrg.bulk_scale_factors(random.uniform(.1, .9)*Pi_min)
    assert nscale > 1 and pscale < 1, \
        'Incorrect scale factors for negative bulk pressure.'

    nscale, pscale = hrg.bulk_scale_factors(-p0)
    assert nscale > 1 and pscale == 0, \
        'Incorrect scale factors for zero total pressure.'

    assert hrg.bulk_scale_factors(-1.1*p0) == (nscale, pscale), \
        'Scale factors should not change outside the Pi range.'

    assert \
        hrg.bulk_scale_factors(Pi_max) == hrg.bulk_scale_factors(1.5*Pi_max), \
        'Scale factors should not change outside the Pi range.'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path
from setuptools import setup, Extension


def version():
    with open('frzout/__init__.py', 'r') as f:
        for l in f:
            if l.startswith('__version__ = '):
                return l.split("'")[1]

    raise RuntimeError('unable to determine version')


def long_description():
    with open('README.rst') as f:
        return f.read()


def ext_modules():
    def module_list(ext):
        return [
            Extension(
                'frzout._frzout',
                ['frzout/_frzout.' + ext],
            ),
            Extension(
                'frzout.test._test_fourvec',
                ['frzout/test/_test_fourvec.' + ext],
                include_dirs=['frzout']
            )
        ]

    # determine release or dev build
    if os.path.exists('PKG-INFO'):
        return module_list('c')
    else:
        try:
            from Cython.Build import cythonize
        except ImportError:
            raise RuntimeError('cython is required for development builds')

        return cythonize(
            module_list('pyx'),
            compiler_directives=dict(cdivision=True, language_level=3)
        )


setup(
    name='frzout',
    version=version(),
    description='Particlization model for relativistic heavy-ion collisions',
    long_description=long_description(),
    author='Jonah Bernhard',
    author_email='jonah.bernhard@gmail.com',
    url='https://github.com/Duke-QCD/frzout',
    license='MIT',
    packages=['frzout', 'frzout.test'],
    package_data={'frzout': ['mass_width_2017.mcd']},
    ext_modules=ext_modules(),
    install_requires=['numpy', 'scipy >= 0.18.0'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Physics'
    ]
)
